import random
import numpy as np

from .solution import Solution
from .types import SwapCandidateList, ReplicateToNeighborDict
from .validity import count_violations, violations_per_replicate
from .util import pairs_to_dict


def find_violating_peptides(
    s: Solution,
) -> tuple[SwapCandidateList, ReplicateToNeighborDict]:
    replicate_to_pool_to_peptides = s.assignments
    # treat invalid pairs as if they've already been neighbors in a previous round
    peptide_to_neighbors = pairs_to_dict(s.invalid_neighbors)
    needs_swap = []
    replicate_to_neighbor_dict = {}
    for replicate_idx, pool_to_peptides in replicate_to_pool_to_peptides.items():
        for pool_idx, peptides in pool_to_peptides.items():
            for p1 in peptides:
                for p2 in peptides:
                    if p1 != p2:
                        if p2 in peptide_to_neighbors[p1]:
                            needs_swap.append((replicate_idx, pool_idx, p2))
                        else:
                            peptide_to_neighbors[p1].add(p2)
        # neighbor constaints at the end of this replicate
        replicate_to_neighbor_dict[replicate_idx] = {
            peptide: neighbors.copy()
            for (peptide, neighbors) in peptide_to_neighbors.items()
        }
    return needs_swap, replicate_to_neighbor_dict


def improve_solution(s: Solution, verbose: bool = False):
    replicate_to_pool_to_peptides = s.assignments

    needs_swap, replicate_to_neighbor_dict = find_violating_peptides(s)

    random.shuffle(needs_swap)

    swapped_pools = set()
    swapped_peptides = set()
    for replicate_idx, pool_idx_a, peptide_a in needs_swap:
        if pool_idx_a in swapped_pools or peptide_a in swapped_peptides:
            continue

        pool_to_peptides = replicate_to_pool_to_peptides[replicate_idx]
        pool_a = pool_to_peptides[pool_idx_a]

        # if a pool is empty, just move the offending peptide there
        empty_pools = {
            pool_idx
            for (pool_idx, pool_peptides) in pool_to_peptides.items()
            if len(pool_peptides) == 0
        }
        if empty_pools:
            pool_idx_b = random.choice(list(empty_pools))
            s.move_peptide(replicate_idx, pool_idx_a, peptide_a, pool_idx_b)
            swapped_pools.add(pool_idx_a)
            swapped_pools.add(pool_idx_b)
            swapped_peptides.add(peptide_a)
        else:
            other_peptides = []
            peptide_to_pool_idx = {}
            for pool_idx_i, pool_peptides_i in pool_to_peptides.items():
                if pool_idx_i != pool_idx_a and pool_idx_i not in swapped_pools:
                    all_peptides_ok = True
                    for p in pool_peptides_i:
                        peptide_to_pool_idx[p] = pool_idx_i
                        all_peptides_ok = all_peptides_ok and (
                            peptide_a
                            not in replicate_to_neighbor_dict[replicate_idx][p]
                        )
                    if all_peptides_ok:
                        other_peptides.extend(
                            [p for p in pool_peptides_i if p not in swapped_peptides]
                        )

            if len(other_peptides) == 0:
                if verbose:
                    print(
                        "Not able to find a valid peptide to swap with for (%s, %s, %s)"
                        % (replicate_idx, pool_idx_a, peptide_a)
                    )
                continue
            peptide_b = random.choice(other_peptides)
            pool_idx_b = peptide_to_pool_idx[peptide_b]
            assert peptide_a != peptide_b
            assert pool_idx_a != pool_idx_b
            pool_b = pool_to_peptides[pool_idx_b]
            if (
                len(pool_a) > 1
                and len(pool_b) < s.max_peptides_per_pool
                and random.choice([False, True])
            ):
                # just move peptide a to the pool with fewer than max peptides
                if verbose:
                    print(
                        "Moving peptide %d from pool %d to pool %d"
                        % (peptide_a, pool_idx_a, pool_idx_b)
                    )
                s.move_peptide(replicate_idx, pool_idx_a, peptide_a, pool_idx_b)
                swapped_pools.add(pool_idx_a)
                swapped_pools.add(pool_idx_b)
                swapped_peptides.add(peptide_a)
            else:
                if verbose:
                    print("Before swap")
                    print(
                        "pool",
                        pool_idx_a,
                        "peptide",
                        peptide_a,
                        pool_to_peptides[pool_idx_a],
                    )
                    print(
                        "pool",
                        pool_idx_b,
                        "peptide",
                        peptide_b,
                        pool_to_peptides[pool_idx_b],
                    )

                # actually swap them
                s.swap_peptides(
                    replicate_idx, pool_idx_a, peptide_a, pool_idx_b, peptide_b
                )

                if verbose:
                    print("After")
                    print(
                        "pool",
                        pool_idx_a,
                        "peptide",
                        peptide_a,
                        pool_to_peptides[pool_idx_a],
                    )
                    print(
                        "pool",
                        pool_idx_b,
                        "peptide",
                        peptide_b,
                        pool_to_peptides[pool_idx_b],
                    )

                swapped_pools.add(pool_idx_a)
                swapped_pools.add(pool_idx_b)
                swapped_peptides.add(peptide_a)
                swapped_peptides.add(peptide_b)


def optimize(
    s: Solution,
    max_iters: int = 100,
    verbose: bool = False,
    add_pool_if_stuck: bool = True,
    return_history: bool = False,
) -> bool:
    """
    Iteratively update solution by randomly swapping a violating peptide with a random other peptide

    Args
    ----
    solution
        Initial solution which will be modified in-place

    max_iters
        Maximum number of swaps to consider performing

    verbose
        Print number of violations for each iteration

    add_pool_if_stuck
        If no improvements have been made for 10 iters, add a pool
        to the last replicate

    return_history
        Return array of constraint validation counts per iteration


    Returns True if non-violating solution found, False if solution still has violations after
    max_iters
    """
    old_num_violations = count_violations(s)
    history = [old_num_violations]
    if verbose:
        print("Initial solution has %s violations" % (old_num_violations,))
    num_iters_without_improvement = 0
    for i in range(max_iters):
        improve_solution(s)
        replicate_to_violation_count = violations_per_replicate(s)
        new_num_violations = sum(replicate_to_violation_count.values())

        history.append(new_num_violations)
        if verbose:
            print("%d) %d -> %d" % (i + 1, old_num_violations, new_num_violations))

        if old_num_violations <= new_num_violations:
            num_iters_without_improvement += 1
        else:
            num_iters_without_improvement = 0

        old_num_violations = new_num_violations

        if new_num_violations == 0:
            if verbose:
                print("Found valid solution after %d swaps" % (i + 1,))
            break

        if num_iters_without_improvement >= 3 and add_pool_if_stuck:
            replicates_in_need = [
                r for (r, v) in replicate_to_violation_count.items() if v > 0
            ]
            assert len(replicates_in_need) > 0
            replicate_idx = min(replicates_in_need)
            s.add_empty_pool(replicate_idx)
            num_iters_without_improvement = 0
            if verbose:
                print(
                    "Adding pool %d to replicate %d"
                    % (len(s.assignments[replicate_idx]), replicate_idx + 1)
                )

    result = old_num_violations == 0

    # just in case we ended up with any empty pools, remove them from the solution
    s.remove_empty_pools()

    if return_history:
        return result, np.array(history)
    else:
        return result



from collections import defaultdict
from typing import Optional, Literal

import numpy as np

from .solution import Solution
from .types import PeptidePairList
from .util import pairs_to_dict, transitive_closure
from .validity import count_violations


def _random_init(
    num_peptides: int = 100,
    peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools: Optional[int] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    verbose: bool = False,
) -> Solution:
    if num_pools is None:
        num_pools = int(np.ceil(num_peptides / peptides_per_pool))

    replicate_to_pool_to_peptides = {}
    for i in range(num_replicates):
        peptide_array = np.arange(num_peptides)
        np.random.shuffle(peptide_array)
        pool_assignments = {}
        replicate_to_pool_to_peptides[i] = pool_assignments

        for j in range(num_pools):
            start_idx = peptides_per_pool * j
            end_idx = peptides_per_pool * (j + 1)
            pool_assignments[j] = peptide_array[start_idx:end_idx]

    return Solution(
        num_peptides=num_peptides,
        max_peptides_per_pool=peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
    )


def _greedy_init(
    num_peptides: int = 100,
    peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools: Optional[int] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    verbose: bool = False,
) -> Solution:
    if num_pools is None:
        num_pools = int(np.ceil(num_peptides / peptides_per_pool))

    peptide_to_invalid = pairs_to_dict(invalid_neighbors)
    peptide_to_preferred = transitive_closure(pairs_to_dict(preferred_neighbors))
    if verbose:
        print("[_greedy_init] Invalid neighbors: %s" % (peptide_to_invalid,))
        print("[_greedy_init] Preferred neighbors: %s" % (peptide_to_preferred,))

    replicate_to_pool_to_peptides = {}

    random_peptide_order = np.arange(num_peptides)
    np.random.shuffle(random_peptide_order)
    assert len(set(random_peptide_order)) == num_peptides
    # assign all peptides with preferred neighbors first
    peptides_with_preferred_neighbors = [
        p for p in random_peptide_order if p in peptide_to_preferred
    ]
    peptides_without_preferred_neighbors = [
        p for p in random_peptide_order if p not in peptide_to_preferred
    ]
    peptide_list = (
        peptides_with_preferred_neighbors + peptides_without_preferred_neighbors
    )
    for i in range(num_replicates):
        peptide_to_pool = {}
        pool_to_peptides = defaultdict(set)

        def add_to_pool(peptide, pool_idx):
            pool = pool_to_peptides[pool_idx]
            pool.add(peptide)
            peptide_to_pool[peptide] = pool_idx
            for other_peptide in pool:
                peptide_to_invalid[peptide].add(other_peptide)
                peptide_to_invalid[other_peptide].add(peptide)

        def curr_num_pools():
            return len(pool_to_peptides)

        def make_new_pool(peptide):
            new_pool_idx = curr_num_pools()
            add_to_pool(peptide, new_pool_idx)
            return new_pool_idx

        for peptide in peptide_list:
            for preferred_neighbor in peptide_to_preferred.get(peptide, []):
                if preferred_neighbor in peptide_to_invalid[peptide]:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d already invalid with preferred neighbor %d"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool_idx = peptide_to_pool.get(preferred_neighbor)
                if preferred_pool_idx is None:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d, preferred neighbor %d not in a pool yet"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool = pool_to_peptides[preferred_pool_idx]
                if len(preferred_pool) >= peptides_per_pool:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, preferred neighbor %d in pool %d which is already full"
                            % (i, preferred_neighbor, preferred_pool_idx)
                        )
                    continue

                add_to_pool(peptide, preferred_pool_idx)
                if verbose:
                    print(
                        "[_greedy_init] replicate %d, adding peptide %d to preferred pool %d to pair with peptide %d"
                        % (i, peptide, preferred_pool_idx, preferred_neighbor)
                    )
                break
            # if we didn't get a preferred peptide pool that's valid
            # and there's room for more pools, just make a singleton
            if peptide not in peptide_to_pool:
                if curr_num_pools() < num_pools:
                    assigned_pool_idx = make_new_pool(peptide)
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, making new pool %d for peptide %d"
                            % (i, assigned_pool_idx, peptide)
                        )

            # otherwise, try to find a valid pool
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        disallowed_peptides = peptide_to_invalid[peptide]
                        valid = all(
                            [
                                other_peptide not in disallowed_peptides
                                for other_peptide in pool
                            ]
                        )
                        if valid:
                            add_to_pool(peptide, pool_idx)
                            break

            # otherwise, pick any pool less than the max size
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        add_to_pool(peptide, pool_idx)
                        break

            # lastly, violate the num_pools constraint to make a
            # new singleton anyways
            if peptide not in peptide_to_pool:
                make_new_pool(peptide)

        replicate_to_pool_to_peptides[i] = {
            pool_idx: np.array(sorted(peptides))
            for (pool_idx, peptides) in pool_to_peptides.items()
        }

    return Solution(
        num_peptides=num_peptides,
        max_peptides_per_pool=peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
    )


def init(
    num_peptides: int = 100,
    peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools: Optional[int] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    strategy: Literal["greedy", "random"] = "greedy",
    verbose=False,
) -> Solution:
    if strategy == "greedy":
        s = _greedy_init(
            num_peptides=num_peptides,
            peptides_per_pool=peptides_per_pool,
            num_replicates=num_replicates,
            num_pools=num_pools,
            invalid_neighbors=invalid_neighbors,
            preferred_neighbors=preferred_neighbors,
            verbose=verbose,
        )
    elif strategy == "random":
        s = _random_init(
            num_peptides=num_peptides,
            peptides_per_pool=peptides_per_pool,
            num_replicates=num_replicates,
            num_pools=num_pools,
            invalid_neighbors=invalid_neighbors,
            preferred_neighbors=preferred_neighbors,
            verbose=verbose,
        )
    else:
        raise ValueError("Unknown initialization strategy: '%s'" % strategy)

    violations = count_violations(s)
    if verbose:
        print("Generated solution with %d initial violations" % (violations))
    return s
